///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoBTreeCursor.cc
// ------------------
// Cego index cursor class implementation
//         
// Design and Implementation by Bjoern Lemke
//     
// (C)opyright 2000-2017 Bjoern Lemke
//
// IMPLEMENTATION MODULE
//
// Class: CegoBTreeCursor
// 
// Description: Cursor class for traversing btree objects
//
// Status: CLEAN
//
///////////////////////////////////////////////////////////////////////////////

// POSIX INCLUDES
#include <string.h>
#include <stdlib.h>

// CEGO INCLUDES
#include "CegoBTreeCursor.h"
#include "CegoTupleState.h"
#include "CegoBTreeNode.h"
#include "CegoBTreeManager.h"

CegoBTreeCursor::CegoBTreeCursor()
{
}

CegoBTreeCursor::CegoBTreeCursor(CegoTableManager *pTabMng, int tabSetId, const Chain& btreeName, CegoObject::ObjectType type, CegoAttrCond* pAttrCond, bool ignoreTouched, bool readUncommitted)
{
    _pTabMng = pTabMng;
    _pDBMng = pTabMng->getDBMng();
    _btreeName = btreeName;
    _type = type;
    _pAttrCond = pAttrCond;

    _tabSetId = tabSetId;
    _ignoreTouched = ignoreTouched;

    _readUncommitted = readUncommitted;
    _cursorCached = false;
    _eoc=true;
    _lockId = 0;
}

CegoBTreeCursor::~CegoBTreeCursor()
{
    abort();
}

bool CegoBTreeCursor::getFirst(ListT<CegoField>& fl, CegoDataPointer& dp)
{
    _eoc=false;

    if ( fl.isEmpty() )
    {
	throw Exception(EXLOC, "Empty field list");
    }

    if ( _cursorCached == false )
    {
	CegoBTreeObject btoe;

	_pTabMng->getObject(_tabSetId, _btreeName, _type, btoe);

	_btreeSchema = btoe.getSchema();

	_keyLen = CegoBTreeValue::getKeyLen(_btreeSchema);

	_rootPageId = btoe.getDataPageId();
	
	Chain tabName = btoe.getTabName();
	
	_cursorCached = true;

	// CegoBTreeManager btm(_pTabMng);
	// btm.dumpBTree(btoe);
    }

    _lockId = _pTabMng->getLockHandler()->lockData(CegoObject::BTREE, _rootPageId, CegoLockHandler::READ);
    
    bool found = false;

    if (_pAttrCond)
    { 
	if ( _pAttrCond->getPrimaryCompMode() == CegoAttrComp::BTWN )
	    _pAttrCond->setPrimaryComparison( MORE_EQUAL_THAN );

	switch ( _pAttrCond->getPrimaryComparison() )
	{
	case NOT_EQUAL:
	case LESS_THAN:
	case LESS_EQUAL_THAN:
	{
	    // go to the beginning

	    bool leafFound = false;
	    _curPageId = _rootPageId;

	    CegoBTreeValue iv;

	    while ( leafFound == false )
	    {
		CegoBufferPage bp;
		_pDBMng->bufferFix(bp, _tabSetId, _curPageId, CegoBufferPool::SYNC, _pTabMng->getLockHandler());
				
		if ( bp.getType() == CegoBufferPage::BTREE_NODE )
		{		    
		    CegoBTreeNode traceNode;
		    traceNode.setPtr(bp.getChunkEntry(), bp.getChunkLen());	    
		    traceNode.setSchema(CegoBTreeNode::NODE, &_btreeSchema, _keyLen);
		    traceNode.setPageId(bp.getPageId());
		    
		    traceNode.reset();
		    traceNode.nextChildPointer(_curPageId);
		    
		    _parentPageStack.Push(bp);
		}
		else if ( bp.getType() == CegoBufferPage::BTREE_LEAF )
		{
		    _curPage = bp;
		    _curLeaf.setPtr(_curPage.getChunkEntry(), _curPage.getChunkLen());
		    _curLeaf.setPageId(_curPage.getPageId());
		    _curLeaf.setSchema(CegoBTreeNode::LEAF, &_btreeSchema, _keyLen);

		    leafFound=true;

		    _curLeaf.reset();

		    bool goOn = true;
		    while ( goOn )
		    {
			CegoBTreeValue iv;
			bool hasValue = _curLeaf.nextValue(iv, dp);
					
			if ( hasValue )
			{	
			    CegoComparison comp = compValue(iv);
			    
			    if ( comp == LESS_THAN
				 || ( comp == EQUAL && _pAttrCond->getPrimaryComparison() == LESS_EQUAL_THAN )
				 || ( comp != EQUAL && _pAttrCond->getPrimaryComparison() == NOT_EQUAL ) )
			    {	       
				found = true;
				goOn = false;
			    }
			    else if ( _pAttrCond->getPrimaryComparison() == NOT_EQUAL )
			    {
				return getNext(fl, dp);
			    }
			}
			else
			{   
			    while ( hasValue == false && goOn )
			    {				
				PageIdType pageId = _curPage.getNextPageId();
								
				_pDBMng->bufferUnfix(_curPage, false, _pTabMng->getLockHandler());
				
				if ( pageId )
				{
				    _pDBMng->bufferFix(_curPage, _tabSetId, pageId, CegoBufferPool::SYNC, _pTabMng->getLockHandler());
				    
				    _curLeaf.setPtr(_curPage.getChunkEntry(), _curPage.getChunkLen());
				    _curLeaf.setPageId(_curPage.getPageId());
				    _curLeaf.setSchema(CegoBTreeNode::LEAF, &_btreeSchema, _keyLen);
				
				    _curLeaf.reset();
				    hasValue = _curLeaf.numEntries() > 0;	    
				}
				else
				{
				    goOn = false;
				}
			    }
			}
		    }	    
		}   
	    }
	    break;
	}
	case EQUAL:
	case MORE_THAN:
	case MORE_EQUAL_THAN:
	{	    
	    bool leafFound = false;

	    _curPageId = _rootPageId;

	    while ( leafFound == false )
	    {
		CegoBufferPage bp;

		_pDBMng->bufferFix(bp, _tabSetId, _curPageId, CegoBufferPool::SYNC, _pTabMng->getLockHandler());

		if ( bp.getType() == CegoBufferPage::BTREE_NODE )
		{	    
		    CegoBTreeNode traceNode;
		    traceNode.setPtr(bp.getChunkEntry(), bp.getChunkLen());	    
		    traceNode.setSchema(CegoBTreeNode::NODE, &_btreeSchema, _keyLen);
		    traceNode.setPageId(bp.getPageId());


		    /* The new implementation uses a more efficient trace algorithm to find the
		       appropriate subnode. The variables lb and rb mark the current boundaries,
		       where the mean value is calculated and checked
		    */
			
		    int lb = 0;
		    int rb = traceNode.numEntries();
		    int pos;
		    
		    bool goOn = true;
		    
		    while ( goOn )
		    {	
			pos = traceNode.getMedPage(lb, rb, _curPageId);		
			
			if ( rb - lb <= 1 )
			{
			    if ( lb != rb )
			    {
				
				// PageIdType leftPageId;
				// pos = traceNode.getMedPage(lb, lb, leftPageId);

				// pos is already set up for left side
				
				CegoBTreeValue ivLeft;
				if ( traceNode.rightValue(pos, ivLeft) )
				{					
				    CegoComparison comp = compValue(ivLeft);
				    
				    if ( comp == EQUAL || comp == MORE_THAN ) 
				    {
					// _curPageId = leftPageId;
				    }				    
				    else
				    {
					// retrieving right pageid		
					PageIdType rightPageId;
					pos = traceNode.getMedPage(rb, rb, rightPageId);
					_curPageId = rightPageId;
				    }				    
				}
			    }			    
			    goOn = false;			    
			}
			else
			{			    
			    // cout  << "BTreeCursor : Medium position is " << pos << endl;
			    // cout  << "BTreeCursor : Current page is " << _curPageId << endl;
				
			    CegoBTreeValue ivRight;
			    
			    if ( traceNode.rightValue(pos, ivRight) )
			    {
				// cout  << "BTreeCursor : Right value is " << ivRight.toChain(&_btreeSchema) << endl;
				CegoComparison comp = compValue(ivRight);
				
				if ( comp == MORE_THAN )
				{	  
				    rb = pos;
				}
				else if ( comp == LESS_THAN )
				{
				    lb = pos;
				}
				else if ( comp == EQUAL )
				{
				    if ( _pAttrCond->getPrimaryComparison() == MORE_THAN )
					lb = pos;
				    else
					rb = pos;
				}
				
				// cout  << "BTreeCursor : lb = " << lb << " rb = " << rb << endl;
			    }
			    else
			    {
				// right end reached, pageId points to next node or leaf
				goOn = false;
			    }			    
			}
		    }
		    
		    /* The following legacy code trace the node with a linear search
		       This is more simple, but also takes linear efforts 
		       ( instead of logarithmic efforts as the implementation above */


		    /*

		    PageIdType checkPageId;
		      
		    traceNode.reset();
		    
		    goOn = traceNode.nextChildPointer(checkPageId);
		    while ( goOn )
		    {
			CegoBTreeValue iv;
			if ( traceNode.rightChild(iv) )
			{
			    CegoComparison comp = compValue(iv);
			    
			    if ( ( comp == EQUAL && _pAttrCond->getPrimaryComparison() != MORE_THAN ) 
				 || comp == MORE_THAN )
			    {	       
				goOn = false;
			    }
			    else
			    {
				goOn = traceNode.nextChildPointer(checkPageId);
			    }
			}
			else
			{
			    goOn = false;
			}			
		    }

		    if ( _curPageId != checkPageId )
		    {
			
			cout << "PAGE MISMATCH" << endl;
			cout << "Valid PageId is " << checkPageId << endl;
			cout << "Wrong PageId is " << _curPageId << endl;
			cout << "lb = " << lb << endl;
			cout << "rb = " << rb << endl;


			lb = 0;
			rb = traceNode.numEntries();
			
			goOn = true;
			
			while ( goOn )
			{	
			    pos = traceNode.getMedPage(lb, rb, _curPageId);		
			    
			    if ( rb - lb <= 1 )
			    {

				if ( lb != rb )
				{
				    
				    // PageIdType leftPageId;
				    // pos = traceNode.getMedPage(lb, lb, leftPageId);
				    
				    // pos is already set up for left side
				    
				    CegoBTreeValue ivLeft;
				    if ( traceNode.rightValue(pos, ivLeft) )
				    {					
					CegoComparison comp = compValue(ivLeft);
					
					if ( comp == EQUAL || comp == MORE_THAN ) 
					{
					    // _curPageId = leftPageId;
					}
					else
					{
					    // retrieving right pageid		
					    PageIdType rightPageId;
					    pos = traceNode.getMedPage(rb, rb, rightPageId);
					    _curPageId = rightPageId;
					}				    
				    }
				}			    
				goOn = false;			    
			    }
			    else
			    {			    
				cout  << "BTreeCursor : Medium position is " << pos << endl;
				cout  << "BTreeCursor : Current page is " << _curPageId << endl;
				
				CegoBTreeValue ivRight;
			    
				if ( traceNode.rightValue(pos, ivRight) )
				{
				    cout  << "BTreeCursor : Right value is " << ivRight.toChain(&_btreeSchema) << endl;
				    CegoComparison comp = compValue(ivRight);
				
				    if ( comp == MORE_THAN )
				    {	  
					rb = pos;
				    }
				    else if ( comp == LESS_THAN )
				    {
					lb = pos;
				    }
				    else if ( comp == EQUAL )
				    {
					if ( _pAttrCond->getPrimaryComparison() == MORE_THAN )
					    lb = pos;
					else
					    rb = pos;
				    }
				    
				    cout  << "BTreeCursor : lb = " << lb << " rb = " << rb << endl;
				}
				else
				{
				    // right end reached, pageId points to next node or leaf
				    goOn = false;
				}			    
			    }
			}
			
			
			traceNode.printNode();
		    }
		    */
		    
		    
		    _parentPageStack.Push(bp);

		}
		else if ( bp.getType() == CegoBufferPage::BTREE_LEAF )
		{		
		    _curPage = bp;
		    _curLeaf.setPtr(_curPage.getChunkEntry(), _curPage.getChunkLen());
		    _curLeaf.setPageId(_curPage.getPageId());
		    _curLeaf.setSchema(CegoBTreeNode::LEAF, &_btreeSchema, _keyLen);

		    leafFound=true;
		    
		    _curLeaf.reset();

		    bool goOn = true;
		    while ( goOn )
		    {
			CegoBTreeValue iv;
			bool hasValue = _curLeaf.nextValue(iv, dp);
	       	
			if ( hasValue )
			{
			    CegoComparison comp = compValue(iv);			
			    if ( ( comp == EQUAL && _pAttrCond->getPrimaryComparison() != MORE_THAN )|| comp == MORE_THAN )
			    {	       
				goOn = false;			    
				if ( comp == MORE_THAN && _pAttrCond->getPrimaryComparison() == EQUAL )
				    found = false;
				else
				    found = true;
			    }			    
			}
			else
			{			    
			    while ( hasValue == false && goOn )
			    {								
				PageIdType pageId = _curPage.getNextPageId();
				
				// cout << "Skipping leave, going to next " << pageId << endl;
			
				_pDBMng->bufferUnfix(_curPage, false, _pTabMng->getLockHandler());
			
				if ( pageId )
				{
				    _pDBMng->bufferFix(_curPage, _tabSetId, pageId, CegoBufferPool::SYNC, _pTabMng->getLockHandler());
				    
				    _curLeaf.setPtr(_curPage.getChunkEntry(), _curPage.getChunkLen());
				    _curLeaf.setPageId(_curPage.getPageId());
				    _curLeaf.setSchema(CegoBTreeNode::LEAF, &_btreeSchema, _keyLen);
				    
				    _curLeaf.reset();
				    hasValue = _curLeaf.numEntries() > 0;
				}
				else
				{
				    goOn = false;
				}
			    }
			}
		    }
		}
	    }	    	    
	    break;	    
	}
	}
    }
    else
    {
	// go to the beginning
	
	bool leafFound = false;
	_curPageId = _rootPageId;
	
	CegoBTreeValue minVal;
	
	while ( leafFound == false )
	{	    
	    CegoBufferPage bp;
	    
	    _pDBMng->bufferFix(bp, _tabSetId, _curPageId, CegoBufferPool::SYNC, _pTabMng->getLockHandler());
	    
	    if ( bp.getType() == CegoBufferPage::BTREE_NODE )
	    {
		CegoBTreeNode traceNode;
		traceNode.setPtr(bp.getChunkEntry(), bp.getChunkLen());	    
		traceNode.setSchema(CegoBTreeNode::NODE, &_btreeSchema, _keyLen);
		traceNode.setPageId(bp.getPageId());
				    	
		traceNode.reset();
		traceNode.nextChildPointer(_curPageId);

		_parentPageStack.Push(bp);

	    }
	    else if ( bp.getType() == CegoBufferPage::BTREE_LEAF )
	    {	
		_curPage = bp;
		_curLeaf.setPtr(bp.getChunkEntry(), bp.getChunkLen());
		_curLeaf.setPageId(bp.getPageId());
		_curLeaf.setSchema(CegoBTreeNode::LEAF, &_btreeSchema, _keyLen);

		CegoBTreeValue iv;
		
		_curLeaf.reset();
		found = _curLeaf.nextValue(iv, dp);
		
		leafFound=true;
	    }	   	    
	}
    }

    if ( found )
    {
	char* p;
	int len;
	
	_pTabMng->releaseDataPtrUnlocked(_dataPtr, false);
	_dataPtr = _pTabMng->claimDataPtrUnlocked(_tabSetId, CegoBufferPool::SYNC, dp, p, len);
	
	// skipping tid
	
	unsigned long long tid;
	unsigned long long tastep;
	CegoTupleState ts;

	int toff = _qh.decodeTupleHeader(tid, tastep, ts, p);

	char* tp = p + toff;
	int tlen = len - toff;
	
	if (tid != 0)
	{	    
	    if ( _ignoreTouched )
	    {			
		if ( ts == INSERTED 
		     && tid == _pTabMng->getTID(_tabSetId) 
		     && tastep < _pTabMng->getTAStep(_tabSetId) )
		{
		    _qh.decodeFVL(fl, tp, tlen);
		    return true;   
		}
		else
		{
		    return getNext(fl,dp);
		}
	    }
	    else
	    {
		if ( ( _readUncommitted == true 
		       && ts == INSERTED )
		     || ( _readUncommitted == false 
			  && (( ts == INSERTED && tid == _pTabMng->getTID(_tabSetId)) 
			      || ( ts == DELETED && tid != _pTabMng->getTID(_tabSetId)))))
		{
		    _qh.decodeFVL(fl, tp, tlen);
		    return true;   
		}
		else
		{
		    return getNext(fl,dp);		    
		}
	    }
	}
	else
	{
	    _qh.decodeFVL(fl, tp, tlen);
	    return true;
	}
    }
    else
    {
	_eoc=true;
    }   
    return false;
}

bool CegoBTreeCursor::getNext(ListT<CegoField>& fl, CegoDataPointer& dp)
{
    if ( _eoc )
	return false;

    if ( fl.isEmpty() )
    {
	throw Exception(EXLOC, "Empty field list");
    }

    CegoBTreeValue iv;
    
    bool moreValue = _curLeaf.nextValue(iv, dp);

    while ( moreValue == false )
    {
	PageIdType pageId = _curPage.getNextPageId();

	_pDBMng->bufferUnfix(_curPage, false, _pTabMng->getLockHandler());

	if ( pageId )
	{
	    _pDBMng->bufferFix(_curPage, _tabSetId, pageId, CegoBufferPool::SYNC, _pTabMng->getLockHandler());
	    
	    _curLeaf.setPtr(_curPage.getChunkEntry(), _curPage.getChunkLen());
	    _curLeaf.setPageId(_curPage.getPageId());
	    _curLeaf.setSchema(CegoBTreeNode::LEAF, &_btreeSchema, _keyLen);
	        
	    _curLeaf.reset();
	    moreValue = _curLeaf.nextValue(iv, dp);
	}
	else
	{
	    _eoc = true;
	    return false;	    
	}
    }

    do {

	bool attrMatch = false;

	// get next available entry from leaf page
        
	if (_pAttrCond)
	{
	    CegoComparison comp = compValue(iv);
	    
	    if ( _pAttrCond->getPrimaryCompMode() == CegoAttrComp::BTWN )
	    {
		_pAttrCond->setPrimaryComparison(LESS_EQUAL_THAN);
	    }
	  	  
	    switch (_pAttrCond->getPrimaryComparison())
	    {
		
	    case EQUAL:
	    {	  
		switch (comp)
		{
		case EQUAL:
		    attrMatch=true;
		    break;
		default:
		    break;
		}
		break;
	    }
	    case NOT_EQUAL:
	    {
		switch (comp)
		{
		case LESS_THAN:
		    attrMatch=true;
		    break;
		case MORE_THAN:
		    attrMatch=true;
		    break;
		case EQUAL:
		    return getNext(fl, dp);
		default:
		    break;
		}
		break;
	    }
	    case LESS_THAN:
	    {		
		switch (comp)
		{
		case LESS_THAN:
		    attrMatch=true;
		    break;
		default:
		    break;
		}
		break;
	    }
	    case MORE_THAN:	     
	    {
		switch (comp)
		{
		case MORE_THAN:
		    attrMatch=true;
		    break;
		default:
		    break;		
		}	
		break;       
	    }
	    case LESS_EQUAL_THAN:
	    {
		switch (comp)
		{
		case EQUAL:
		    attrMatch=true;
		    break;
		case LESS_THAN:
		    attrMatch=true;
		    break;
		default:
		    break;
		}
		break;
	    }   
	    case MORE_EQUAL_THAN:
	    {
		switch (comp)
		{
		case EQUAL:
		    attrMatch=true;
		    break;
		case MORE_THAN:
		    attrMatch=true;
		    break;
		default:
		    break;
		}
		break;
	    }
	    }	    
	}
	else
	{
	    attrMatch = true;
	}
	
	if ( attrMatch )
	{
	    
	    char* p;
	    int len;
	    	    
	    // dp = _ie.getData();
	    
	    _pTabMng->releaseDataPtrUnlocked(_dataPtr, false);
	    _dataPtr = _pTabMng->claimDataPtrUnlocked(_tabSetId, CegoBufferPool::SYNC, dp, p, len);
	    
	    // skipping tid
	    
	    unsigned long long tid;
	    unsigned long long tastep;
	    CegoTupleState ts;

	    int toff = _qh.decodeTupleHeader(tid, tastep, ts, p);

	    char* tp = p + toff;
	    int tlen = len - toff;

	    if (tid != 0)
	    {		    
		if ( _ignoreTouched )
		{
		    if ( ts == INSERTED 
			 && tid == _pTabMng->getTID(_tabSetId) 
			 && tastep < _pTabMng->getTAStep(_tabSetId) )
		    {
			_qh.decodeFVL(fl, tp, tlen);
			return true;   
		    }
		    else
		    {
			return getNext(fl, dp);		
		    }
		}
		else
		{
		    if ( ( _readUncommitted == true 
			   && ts == INSERTED )
			 || ( _readUncommitted == false 
			      && (( ts == INSERTED && tid == _pTabMng->getTID(_tabSetId)) 
				  || ( ts == DELETED && tid != _pTabMng->getTID(_tabSetId)))))
		    {
			_qh.decodeFVL(fl, tp, tlen);
			return true;
		    }
		    else
		    {
			return getNext(fl, dp);		
		    }
		}
	    }
	    else
	    {	       
		_qh.decodeFVL(fl, tp, tlen);
		return true;
	    }
	}
	else
	{
	    abort();
	    _eoc = true;
	    return false;
	}
    }  while (1);
}

void CegoBTreeCursor::abort()
{
    _pTabMng->releaseDataPtrUnlocked(_dataPtr, false);
    _dataPtr = CegoBufferPage();
	
    if ( _curPage.isFixed() )
	_pDBMng->bufferUnfix(_curPage, false, _pTabMng->getLockHandler());	

    // unfix node pages on stack
    CegoBufferPage bp;
    while ( _parentPageStack.Pop(bp) )
    {
	// cout << "Unfixing parent page " << bp.getFileId() << " " << bp.getPageId() << endl;
	_pDBMng->bufferUnfix(bp, false, _pTabMng->getLockHandler());
    }

    if ( _lockId )
    {
	_pTabMng->getLockHandler()->unlockData(CegoObject::BTREE, _lockId);
	_lockId = 0;
    }
}

void CegoBTreeCursor::reset()
{
    abort();
}

CegoComparison CegoBTreeCursor::compValue(const CegoBTreeValue& val)
{

    ListT<CegoFieldValue> fvl = val.valueToFVL(_btreeSchema);

    CegoField* pF = _btreeSchema.First();
    CegoFieldValue *pFV = fvl.First();

    while ( pF ) 
    {
	CegoAttrComp* pAC = _pAttrCond->getAttrCompSet().First();

	while ( pAC )
	{
	    if  ( (Chain)pAC->getAttrName() == (Chain)pF->getAttrName()  )
	    {	       
		if ( *pFV < pAC->getFieldValue() )
		{
		    return LESS_THAN;
		}		
		else if ( pAC->getCompMode() == CegoAttrComp::VAL
		     || pAC->getCompMode() == CegoAttrComp::ATTR )
		{
		    if ( *pFV > pAC->getFieldValue() )
		    {
			return MORE_THAN;
		    }
		}
		else if ( pAC->getCompMode() == CegoAttrComp::BTWN )
		{
		    if ( *pFV > pAC->getFieldValue2() )
		    {
			return MORE_THAN;
		    }
		}		
	    }
	    pAC = _pAttrCond->getAttrCompSet().Next();

	    // check for partial modus here
	    if ( pAC )
	    {
		if ( pAC->getComparison() != EQUAL )
		    return EQUAL;
	    }
	}
	pF = _btreeSchema.Next();
	pFV = fvl.Next();
    }
    return EQUAL;
}
